module HarmTrace.Audio.Evaluation where

import HarmTrace.Audio.ChordTypes
import HarmTrace.Audio.Utils 
import HarmTrace.Audio.Key(getBeatSyncKeyFromChroma)
import HarmTrace.Audio.BeatChroma ( createChordRanks, groupPChord, syncEndings)
import HarmTrace.Base.MusicRep 

import Data.List (genericLength, zipWith5)
import Text.Printf(printf)
import System.IO (stdout,hFlush)

--------------------------------------------------------------------------------
-- Evaluation Parameters
--------------------------------------------------------------------------------

-- this functions determines when two chords are considered the same
eqFunc :: ChordLabel -> ChordLabel -> Bool
eqFunc = chordTriadEq

sampleRate, displaySampleRate :: NumData
-- the sample rate used in a normal (non-visual) comparison (in seconds)
sampleRate        = 0.01
-- the sample rate used when visually comparing a chord annotation with a 
-- ground-truth annotation. Often a higher sample rate is prefered. Although
-- one uses precision, the visual result is easier to read
displaySampleRate = 0.3

--------------------------------------------------------------------------------
-- Evaluation function
--------------------------------------------------------------------------------

chordClassEq :: ChordLabel -> ChordLabel -> Bool
chordClassEq (Chord (Note Nothing N) None _ _ _)
             (Chord (Note Nothing N) None _ _ _)       = True
chordClassEq (Chord (Note Nothing N) None _ _ _) _     = False
chordClassEq _     (Chord (Note Nothing N) None _ _ _) = False
chordClassEq (Chord r1 sh1 _ _ _) (Chord r2 sh2 _ _ _) =
  (diaNatToSemi r1) == (diaNatToSemi r2) && (toClassType sh1 == toClassType sh2)

chordTriadEq :: ChordLabel -> ChordLabel -> Bool
chordTriadEq (Chord (Note Nothing N) None _ _ _)
             (Chord (Note Nothing N) None _ _ _)       = True
chordTriadEq (Chord (Note Nothing N) None _ _ _) _     = False
chordTriadEq _     (Chord (Note Nothing N) None _ _ _) = False
chordTriadEq (Chord r1 sh1 _ _ _) (Chord r2 sh2 _ _ _) =
  (diaNatToSemi r1) == (diaNatToSemi r2) && (toMode sh1 == toMode sh2)

-- calculates the relative correct overlap, which is the recall
-- of matching frames, and defined as the nr of matching frames (sampled at
-- an 10 milisecond interval) divided by all frames
relCorrectOverlap :: ChordAnnotation -> ChordAnnotation -> Double
relCorrectOverlap a b =
    (foldl countMatch 0 (zipWith eqFunc sama samb)) / tot where
    sama = sample a
    samb = sample b
    tot  = max (genericLength sama) (genericLength samb)

-- does the same thing as relCorrectOverlap, but it also prints the
-- chords and uses a lower sample rate
printRelCorrectOverlap :: (AudioFeat -> ChordAnnotation) -> AudioFeat
                       -> ChordAnnotation -> IO Double
printRelCorrectOverlap annotator af@(AudioFeat _idStr chrm' beat' key') gt = do
  let (chrm, beat, key) = syncEndings chrm' beat' key'
      keys  = mergeAndTimeStamp head  beat $ getBeatSyncKeyFromChroma beat key   
      blks  = mergeAndTimeStamp avgPC beat . groupPChord
            . createChordRanks $ beatSync beat chrm
      -- sample the info for printing and evaluation
      samaf = sampleWith displaySampleRate (annotator af)
      samgt = sampleWith displaySampleRate gt
      sambk = sampleWith displaySampleRate blks
      samk  = sampleWith displaySampleRate keys

      tot   = max (genericLength samaf) (genericLength samgt)
      showEq m = if m then "==" else "/=" 
      printEval :: NumData -> ChordLabel -> ChordLabel -> Key -> [ProbChord] 
                -> IO Bool
      printEval t g a b c = 
         do putStrLn (printf "%.2f" t ++ '\t' : showEq equal ++ '\t' : show g
                      ++ '\t' : show a ++ '\t' : show b ++ '\t' : show c)
                      >> hFlush stdout
            return equal where equal = g `eqFunc` a
  putStrLn "time\tmatch\tGT\t\tMPTREE\tkey\toptional chords"
  m <- sequence (zipWith5 printEval [0.0,displaySampleRate ..] 
                                    samgt samaf samk sambk)
  return ((foldl countMatch 0 m) / tot)

countMatch :: Double -> Bool -> Double 
countMatch x y | y         = succ x -- count the number of matching frames
               | otherwise = x

-- given a chord annotation sample the chord label at every 10 ms
sample :: [TimedData a] -> [a]
sample = sampleWith sampleRate

-- like sample, but takes a sample rate (seconds :: Float) as argument
sampleWith :: NumData -> [TimedData a] -> [a]
sampleWith rate =  sampleAt [0.00, rate .. ] 
