{-# OPTIONS_GHC -Wall                     #-}
{-# LANGUAGE GADTs                        #-}
{-# LANGUAGE ScopedTypeVariables          #-}
{-# LANGUAGE TupleSections                #-}

module HarmTrace.Audio.Harmonize ( harmonyAnnotator, headAnnotator
                                 , simpleAnnotator , putSegStats) where

import HarmTrace.Audio.BeatChroma( createChordRanks, syncWithAnnKey, addTimeInfo
                                 , groupPChord, segmentByKey , syncEndings)
import HarmTrace.Audio.Key (getBeatSyncKeyFromChroma)
import HarmTrace.Audio.Utils 
import HarmTrace.Audio.ChordTypes
import HarmTrace.Base.MusicRep

import HarmTrace.Models.Models
import HarmTrace.Models.Jazz.Main
import HarmTrace.Models.Pop.Main
import HarmTrace.Tokenizer.Tokens
import HarmTrace.IO.Errors

import HarmTrace.HAnTree.HAn (HAn)
import HarmTrace.HAnTree.Tree (Tree, size, depth)
import HarmTrace.HAnTree.ToHAnTree (GTree)
-- import HarmTrace.HAnTree.PostProcess (PPOption (..))
import HarmTrace.HarmTrace

import Text.ParserCombinators.UU
import Text.ParserCombinators.UU.BasicInstances

import System.IO (stdout,hFlush)
import Data.List (sortBy, groupBy)
import Text.Printf (printf)

--------------------------------------------------------------------------------
-- Local Parameters
--------------------------------------------------------------------------------

maxSegmentSize, maxLProductSize :: Int
maxSegmentSize  = 8
maxLProductSize = 20

--------------------------------------------------------------------------------
-- From chords with probabilities to a single chord, using harmony
--------------------------------------------------------------------------------
       
harmonyAnnotator :: GrammarEx -> Maybe [TimedData Key] -> AudioFeat 
                 -> ChordAnnotation
harmonyAnnotator (GrammarEx g) mk af = case mk of 
  Nothing  -> concatMap (harmonize g) (preProcessData af)
  (Just k) -> concatMap (harmonize g) (preProcessDataWithKey k af)

preProcessDataWithKey :: [TimedData Key] -> AudioFeat -> [ProbChordSeg]
preProcessDataWithKey key' (AudioFeat _idStr chrm beat' _keystrength) = 
  let (beat,key) = syncWithAnnKey chrm beat' key'
      chdgrp     = mergeAndTimeStamp avgPC beat . groupPChord
                 . createChordRanks $ beatSync beat chrm
  in segmentByTonic $ segmentByKey key chdgrp
  
preProcessData :: AudioFeat -> [ProbChordSeg]
preProcessData (AudioFeat _idStr chrm' beat' key') = 
  let (chrm, beat, key) = syncEndings chrm' beat' key'
      keygrp = mergeAndTimeStamp head beat $ getBeatSyncKeyFromChroma beat key   
      chdgrp = mergeAndTimeStamp avgPC beat . groupPChord
             . createChordRanks $ beatSync beat chrm
  in segmentByTonic $ segmentByKey keygrp chdgrp

harmonize :: forall g. (GTree g) => Grammar g -> ProbChordSeg -> ChordAnnotation
harmonize g (Segment k bs) =
  let isExpandable :: Bool
      isExpandable = (length $ filter ((>1) . length) (map getData bs)) > 0
      
      pickHead :: Block -> ChordSegment
      pickHead tpc = tpc {getData = chordLab . head $ getData tpc}
      
      myParse :: [ChordToken] -> (Tree HAn,[ChordToken],[Error Int])
      myParse x =
        let -- First, parse the tokens
            res :: ([g],[Error Int])
            res = case g of 
                      Jazz -> parse_h ((,) <$> pJazz k <*> pEnd) (createStr 0 x)
                      Pop  -> parse_h ((,) <$> pPop  k <*> pEnd) (createStr 0 x)
            -- Build a ParseResult from that
            pr = ParseResult u (concatMap chords x) (fst res) u u u (snd res) []
            -- So that we can post-process it. Then extract the Tree HAn
            t  = pieceTreeHAn (postProc [ RemovePDPT, MergeDelChords ]   pr)
            u :: forall a. a
            u  = error "harmonize: undefined placeholder evaluated"
        -- Return the Tree HAn, the input tokens, and the errors
        in (t, x, snd res)

      -- To be improved
      evaluateParse :: (Tree HAn,[ChordToken],[Error Int])
                    -> (Tree HAn,[ChordToken],Float)
      evaluateParse (ts,tokens,errors) = (ts,tokens,errorRatio errors tokens)

      -- Generate, parse, and evaluate all possible sequences of chords
      parseResults :: [(Tree HAn,[ChordToken],Float)]
      parseResults = [ evaluateParse (myParse l)
                     | l <- lProduct (map (map probChord . getData) bs) ]

      -- From all possible parse trees, take the best one
      select :: [(Tree HAn,[ChordToken],Float)] -> [ChordToken]
      select = select1 . head
             . groupBy (\(_,_,a) (_,_,b) -> a `compare` b == EQ)
             . sortBy  (\(_,_,a) (_,_,b) -> a `compare` b)

      -- These all have the same error ratio, so we sort them first by tree
      -- size, then depth, and pick the first
      select1 :: [(Tree HAn,[ChordToken],Float)] -> [ChordToken]
      select1 = snd3 . head . sortBy cmp where
        cmp (a,_,_) (b,_,_) = (size a, depth a) `compare` (size b, depth b)
        snd3 (_,a,_) = a

      probChord :: ProbChord -> ChordToken
      probChord (ProbChord lab@(Chord r sh _add _on _dur) _p) = 
        (ChordToken r' sh' [lab] NotParsed 1 0) where
           r'  = if isNone r   then Note Nothing Imp else toScaleDegree k r
           sh' = if sh == None then NoClass          else toClassType sh
      
      -- store on and offsets for merging after parsing
      ons, offs :: [NumData]
      ons  = map onset  bs
      offs = map offset bs

      mergeTimeStamp :: [ChordToken] -> [ChordSegment]
      mergeTimeStamp cs = zipWith3 TimedData (map (head . chords) cs) ons offs

      -- if there is nothing to expand, do not parse
  in if   isExpandable then mergeTimeStamp $ select parseResults 
     else map pickHead bs

--------------------------------------------------------------------------------
-- Segmentation functions
--------------------------------------------------------------------------------
-- Temporary test values
{-
test = segmentTonic testKey testSeq
testKey = Key (Note Nothing C) MajMode
testSeq = testChordG ++ testChordC ++ testChordC ++ testChordG ++ testChordG
testChordC = [TimedData [ProbChord labC 1, ProbChord labG 0.5] 0 0]
testChordG = [TimedData [ProbChord labG 1, ProbChord labC 0.5] 0 0]
labC = Chord (Note Nothing C) Maj [] 0 0
labG = Chord (Note Nothing G) Maj [] 0 0
-}

-- Reminder: TimedData [ProbChord] == Block
segmentByTonic :: [ProbChordSeg] -> [ProbChordSeg]
segmentByTonic segs = concatMap emergencySplit $ concatMap split segs where
  split :: ProbChordSeg -> [ProbChordSeg]
  split (Segment key cs) = zipWith Segment (repeat key) (segmentTonic key cs)

-- In case segments are just to big, even after segmenting on Tonic and Dominant
-- split these segments into smaller segements recursively.
emergencySplit :: ProbChordSeg -> [ProbChordSeg]
emergencySplit (Segment k cs) = map (Segment k) (recSplit cs)
recSplit :: [TimedData [a]] -> [[TimedData [a]]]
recSplit [] = []
recSplit b
  |     blen                <= maxSegmentSize 
    && (snd $ lProdStats b) <= maxLProductSize = [b]
  | otherwise   = recSplit l ++ recSplit r
    where blen  = length b
          (l,r) = splitAt (blen `div` 2) b 
  
-- Break into segments according to the key
segmentTonic :: Key -> [Block] -> [[Block]]
segmentTonic k' cs' = segmentTonic' k' cs' [] where 
  segmentTonic' _ []     []     = []
  segmentTonic' _ []     acc    = [reverse acc]
  segmentTonic' k (c:cs) acc
    | c' `isTonic` k || c' `isDom` k = reverse (c:acc) : segmentTonic k cs
    | otherwise                      = segmentTonic' k cs (c:acc) where 
        c' = getFstChord c

-- Take the first chord (which is the one with the highest probability, since
-- the list is sorted)
getFstChord :: TimedData [ProbChord] -> ChordLabel
getFstChord c = case getData c of
                  []    -> error "getFstChord: empty list"
                  (h:_) -> chordLab h -- only split on chords we are certain of
                -- _   -> Chord (Note Nothing N) None [] 0 0 -- else return None

-- Check if this chord label is the tonic
isTonic :: ChordLabel -> Key -> Bool
isTonic c (Key r m) = r == chordRoot c && m `eqMode` (chordShorthand c)

-- Check if this chord label is the dominant
-- JPM: I don't understand why this function looks so different from `isTonic`
isDom :: ChordLabel -> Key -> Bool
isDom (Chord (Note Nothing N) _ _ _ _) _ = False
isDom c key       =    toScaleDegree key (chordRoot c) == Note Nothing V
                    && MajMode `eqMode` (chordShorthand c)

-- It is debatable how to implement this function, musically speaking. This
-- is what I came up with, without thinking too much.
eqMode :: Mode -> Shorthand -> Bool
eqMode _ Sus4  = False
eqMode _ Sus2  = False
eqMode m sh    = m == toMode sh

lProduct :: [[a]] -> [[a]]
lProduct []    = []
lProduct [l]   = [ [x] | x <- l ]
lProduct (h:t) = concat [ map (x:) (lProduct t) | x <- h ]

--------------------------------------------------------------------------------
-- Some printing and statistics functions
--------------------------------------------------------------------------------

-- prints Segmetation statistics
putSegStats :: Maybe [TimedData Key] -> AudioFeat -> IO()
putSegStats mk af = case mk of
  Nothing  -> mapM_ segmentStat $ preProcessData af
  (Just k) -> mapM_ segmentStat $ preProcessDataWithKey k af 

segmentStat :: ProbChordSeg -> IO ()
segmentStat s@(Segment k bs) = 
  do putStr ("start: " ++ (printf "%.3f" . onset  $ head bs))
     putStr (", end: "  ++ (printf "%.3f" . offset $ last bs))
     putStr (", key: "  ++ show k)
     putStr (", probChords: " ++ show (length bs))
     let (l, lpr) = lProdStats bs
     putStr (", lists > 1: "  ++ (show l) )
     putStrLn (" lProduct: " ++ show lpr)
     print s >> hFlush stdout

-- Given a Block list this function returns the number of probChords with a 
-- list > 1 (fst) and the lProduct size (snd)
lProdStats :: [TimedData [a]] -> (Int, Int)
lProdStats bs = (length l, lpr) where
  l   = filter ((>1) . length ) (map getData bs)
  lpr = foldr (\a b -> (length  a) * b) 1 l

--------------------------------------------------------------------------------
-- A baseline chord label annotator
--------------------------------------------------------------------------------

-- Creates an annotation out of a block list, by just picking the first chord
-- This annotator does smart grouping
headAnnotator :: Maybe [TimedData Key] -> AudioFeat -> ChordAnnotation
headAnnotator _keyAnn (AudioFeat _idStr chrm beat _key ) = -- ignore key info
  let endTime = time $ last chrm
      bts     = takeWhile (< endTime) beat ++ [endTime]
  in map headChrd . mergeAndTimeStamp avgPC bts . groupPChord . createChordRanks 
  $ beatSync bts chrm where 
    headChrd :: TimedData [ProbChord] -> TimedData ChordLabel
    headChrd td = td {getData = chordLab . head $ getData td}

-- The most simple annotator, no grouping, no matching, 
-- just pick the best matching chord
simpleAnnotator :: Maybe [TimedData Key] -> AudioFeat -> ChordAnnotation
simpleAnnotator _keyAnn (AudioFeat _idStr chrm beat _key ) = -- ignore key
  addTimeInfo beat . map (chordLab . head) . createChordRanks
  $ beatSync beat chrm
  